if(file.exists("CHARMhelperFunctions.r")) {
  source("CHARMhelperFunctions.r")
} else {
  cat("  Warning: Could not find file \"CHARMhelperFunctions.r\". \n")
  cat("  Make sure to download it and put it in working directory.\n")
  cat("  If you are manually sourcing each file, then feel free ignore this warning.\n")
}
if(file.exists("mathFns.r")) {
  source("mathFns.r")
} else {
  cat("  Warning: Could not find file \"mathFns.r\". \n")
  cat("  Make sure to download it and put it in working directory.\n")
  cat("  If you are manually sourcing each file, then feel free ignore this warning.\n")
}

# Function to make simple test data for users to get started with
# x = makeTestData()
makeTestData=function(numInds=500, numVars=10, mafs=rep(0.5, numVars))
{
  # Note that CHARM assumes that each row is an individual, while each column is a variant
  indNames = paste("ind", 1:numInds, sep="")
  varNames = paste("var", 1:numVars, sep="")
  genMat   = matrix(rbinom(n=1e3, size=2, p=0.5), ncol=numVars, nrow=numInds, byrow=TRUE)
  rownames(genMat) = indNames
  colnames(genMat) = varNames
  caseCon  = c(rep(0, floor(numInds/2)), rep(1, numInds-floor(numInds/2)))
  names(caseCon) = indNames
  
  testDat  = list("genMat"=genMat,
                 "caseCon"=caseCon)
  return(testDat)
}

# Function to run CHARM on some test data, returning the standard list that CHARM returns
# Example usage:
# CHARMresults = runCHARMonTestData()
runCHARMonTestData=function(testData=NULL)
{
  if(is.null(testData)) {
    cat("  Test data not passed, generating some on the fly.\n")
    testData = makeTestData()
  }
  genMat       = testData$genMat
  caseCon      = testData$caseCon
  CHARMresults = CHARMmain(rGen=genMat, caseControlStatus=caseCon)
  cBF          = exp(CHARMresults[["natLogBayesFactor"]])
  log10cB      = CHARMresults[["Log10BayesFactor"]]
  cat("  \"Cross Validated Bayes Factor\" supporting association = ", 
         round(cBF, 2), " (log10 = ", round(log10cB, 2), ")\n", sep="")
  cat("\n")
  cat("  This can be used as a rough guide, where log10 values above 2 indicate weak evidence,\n", 
      "  and log10 values above 3 indicate moderate evidence, while log10 values above 4 indicate\n",
      "  strong evidence for association. However these are approximate, and perumutation should\n", 
      "  be used for final results.\n", sep="")
  return(CHARMresults)
}




CHARMmain=function(
                           models                = getRegionAssocModels() ,
                           modelChoiceDetails    = getModelChoiceDetails(),
                           crossValDetails       = getCrossValDetails()   ,
                           glmControl            = glm.control()          , 
                          #dPar                  = getDiseaseModelPars()  ,
                           useAverageofLogMLs    = TRUE,
                           rGen,
                           caseControlStatus,
                           DEBUG=1, verbose=DEBUG
                          )
{
  if(verbose>=0.5)
    cat("  Model averaging with ", crossValDetails[["fold"]], " fold cross validations (", 
        crossValDetails[["numRepeatsOfCrossVal"]], " repeats) :", sep="")
  # Calculating the frequencies of the one-alleles (which are then used to get the MAFs):
  OAFs              = colSums(rGen)/(2*nrow(rGen))
  MAFs              = pmin(OAFs, 1-OAFs)
  numCrossValRepeats  = crossValDetails[["numRepeatsOfCrossVal"]]
  natLogBayesFactors  = rep(NA, numCrossValRepeats)
  wholeModelMargLikes = rep(NA, numCrossValRepeats)
  lnullModelLikes     = rep(NA, numCrossValRepeats)
  subModelMargLikeMat = matrix(NA, nrow=numCrossValRepeats, ncol=models[["numModels"]])
  for(thisRepeat in 1:numCrossValRepeats)
  {
    if(numCrossValRepeats>=10) {
      if(verbose>=1 & thisRepeat%%floor(numCrossValRepeats/10)==0)
        cat("*")  
    } else
      if(verbose>=1)
        cat("*")      
    cvrv = CHARMcore(
                           models                = models            ,
                           modelChoiceDetails    = modelChoiceDetails,
                           crossValDetails       = crossValDetails   ,
                           glmControl            = glmControl        ,
                          #dPar                  = dPar              , #This should be obsolete now
                           rGen                  = rGen              ,
                           MAFs                  = MAFs              ,
                           caseControlStatus     = caseControlStatus ,
                           DEBUG=DEBUG, verbose=verbose)
    natLogBayesFactors[ thisRepeat ] = cvrv[["natLogBayesFactor" ]]
    wholeModelMargLikes[thisRepeat ] = cvrv[["wholeModelMargLike"]]
    lnullModelLikes[    thisRepeat ] = cvrv[["lnullModelLike"    ]]
    subModelMargLikeMat[thisRepeat,] = cvrv[["subModelMargLikes" ]]
    if(is.null(colnames(subModelMargLikeMat)))
      colnames(subModelMargLikeMat) = names(cvrv[["subModelMargLikes" ]])
  }
  indivResults = list(
            "natLogBayesFactors"  = natLogBayesFactors ,
            "wholeModelMargLikes" = wholeModelMargLikes,
            "subModelMargLikeMat" = subModelMargLikeMat,
            "lnullModelLikes"     = lnullModelLikes
           )
    if(useAverageofLogMLs) {
    natLogAvgWholeModelMargLikes = mean(wholeModelMargLikes) 
    natLogAvgsubModelMargLikes   = colMeans(subModelMargLikeMat)
    natLogAvgNullModelLikes      = mean(lnullModelLikes) 
  } else {
    # avgLog  takes in a vector, log(a, base=exp(1)) and returns a scalar log(average(a), base=exp(1))
    natLogAvgWholeModelMargLikes = avgLog(wholeModelMargLikes) 
    natLogAvgsubModelMargLikes   = apply(subModelMargLikeMat, MARGIN=2, FUN=avgLog)
    natLogAvgNullModelLikes      = avgLog(lnullModelLikes    ) 
  }
  names(natLogAvgsubModelMargLikes) = models$modelNames
  natLogAvgBayesFactors        = natLogAvgWholeModelMargLikes - natLogAvgNullModelLikes
  if(verbose >= 0.5)
    cat("\n")
  if(verbose>=1.1)
    cat("  Cross validation  approach: *Average* log10(Bayes Factor)      : ", natLogAvgBayesFactors * log10(exp(1)), "\n")
  if(verbose>2)
    print(indivResults$subModelMargLikeMat)
   #print(crossValResults$indivResults$subModelMargLikeMat)
  models = c(models, "useAverageofLogMLs"=useAverageofLogMLs)
  rv = list("natLogBayesFactor"   = natLogAvgBayesFactors       ,
            "Log10BayesFactor"    = natLogAvgBayesFactors*log10(exp(1)),
            "wholeModelMargLike"  = natLogAvgWholeModelMargLikes,
            "subModelMargLikes"   = natLogAvgsubModelMargLikes  ,
            "lnullModelLike"      = natLogAvgNullModelLikes     ,
            "indivResults"        = indivResults                ,
            "numCrossValRepeats"  = numCrossValRepeats          ,
            "modelsUsed"          = models                      ,
            "modelChoiceDetails"  = modelChoiceDetails          ,
            "crossValDetails"     = crossValDetails 
           )
  return(rv)
}








CHARMcore=function(rGen,                 # Matrix of genotypes coded as 0, 1, 2
                           caseControlStatus,    # binary (integer) vector, 0=control, 1=case
                           models                = getRegionAssocModels() ,
                           modelChoiceDetails    = getModelChoiceDetails(),
                           crossValDetails       = getCrossValDetails()   ,
                           glmControl            = glm.control()          , 
                           # dPar                  = getDiseaseModelPars()  , #Should be obsolete now
                           MAFs = NULL,
                           DEBUG=1, verbose=DEBUG
                          )
{
  genCounts  = apply(rGen, 2, table)
  if(is.matrix(genCounts)) {
    missingDat = sum(colSums(genCounts) == nrow(rGen)) != ncol(rGen)
  } else {
    missingDat = sum(sapply(genCounts, sum) == nrow(rGen)) != ncol(rGen)
  }
  if(missingDat) {
    cat("  Looks like there's missing data present. Currently this test cannot handle missing data.\n")
    browser()
  }
  require(arm)

  lnullModelLike = sum(dbinom(x=caseControlStatus, size=1, 
                       prob=mean(caseControlStatus), log=TRUE))

  if(verbose>1)
    cat(" Model averaging with cross-val:")
  numModels    = length(  models[["pNumHits"  ]])
  numCrossVals = crossValDetails[["numRepeats"]]
  crossValFold = crossValDetails[["fold"      ]]
  if(numCrossVals != crossValFold)
    cat("  Warning, for the most natural 'likelihood' would expect to calculate \n",
        "  the test probability exactly once for each individual.\n") 
  numInds      = nrow(rGen)
  numSNPs      = ncol(rGen)
  if(is.null(MAFs)) {
    # Calculating the frequencies of the one-alleles (which are then used to get the MAFs):
    OAFs              = colSums(rGen)/(2*nrow(rGen))
    MAFs              = pmin(OAFs, 1-OAFs)
  }  
    

  myModels = matrix(list(), nrow=numCrossVals, ncol=numModels)
  myModelColnames = models$longModelNames[which(models$modelsToUse==1)]
  colnames(myModels) = myModelColnames

  crossValLikelihoods = matrix(NA, nrow=numCrossVals, ncol=numModels)
  dimnames(crossValLikelihoods) = dimnames(myModels)

  testTrainInds = getTestAndTrainInds(individuals=caseControlStatus, crossValDetails=crossValDetails)

  # Break up the individuals into a set of training and test sets, 
  # first decide how many individuals to have in each block (res holds these breaks)
  isNextInt=function(inVec, index) return(floor(inVec[index+1])==(floor(inVec[index])+1))
  indBreaks = (1:numInds)/(numInds/crossValFold)
  res = sort(unique(c(0, which(sapply(1:(length(indBreaks)-1), FUN=isNextInt, inVec=indBreaks))+1, numInds)))

  for(crossValIndex in 1:numCrossVals)
  {
    if(verbose>1)
      cat(".")
    if(crossValDetails$randomInds)
      randomisedCCSInds = sample((1:numInds), numInds, replace=FALSE) else
      randomisedCCSInds = (1:numInds)
    testInds  = sort(randomisedCCSInds[(res[crossValIndex]+1):res[crossValIndex+1]])
    trainInds = (1:numInds)[-testInds]
    for(modelIndex in 1:numModels) 
    {
      if(!crossValDetails[["replace"]]) {
        trainSubsetRgen = rGen[trainInds, , drop=FALSE]
        testSubsetRgen  = rGen[testInds , , drop=FALSE]
        # Removing monomorphic SNPs from training and testing data. Will give non comformable outputs, for xval likelihoods!!!
        # trainSubsetRgen = trainSubsetRgen[, colSums(trainSubsetRgen)>0,drop=FALSE]
        # testSubsetRgen  = testSubsetRgen[ , colSums(testSubsetRgen )>0,drop=FALSE]
        trainSubsetccs  = caseControlStatus[trainInds]
        testSubsetccs   = caseControlStatus[testInds ]
      } else {
        cat("  Cross val with replacement not implemented.\n")
        browser()
      }
      if(ncol(rGen>0))
        formula = trainSubsetccs~trainSubsetRgen else
        formula = trainSubsetccs~matrix(1, nrow=nrow(trainSubsetRgen))
     #cat("  SNPdf = ", models[["SNPdf"      ]][modelIndex], "...", sep="")
      if(models$MAFdependES[modelIndex]==0)
        scaleType = "bestAprioriGuess" else
        scaleType = "MAFdependent"
      scales = getScales(DEBUG=DEBUG, verbose=verbose-1, usage=scaleType,  
                         SNPdf       =           models[["SNPdf"      ]][modelIndex], 
                         pNumHits    =           models[["pNumHits"   ]][modelIndex], 
                         pMinHitSize = RRtoOR(RR=models[["pMinHitSize"]][modelIndex], 
                                              prev = models$prev),
                         MAFs        = MAFs,
                         numSNPReg   = numSNPs)
     #browser()
      warn = options()$warn
      options(warn=0)
      myModels[[crossValIndex, modelIndex]] = 
        eval(bayesglm(
                  formula                   = formula                      ,
                  family                    = binomial(link = "logit")     ,
                  prior.mean                = 0                            ,
                  prior.scale               = scales                      ,
                  prior.df                  = models[["SNPdf"]][modelIndex],
                  prior.mean.for.intercept  = 0                            ,
                  prior.scale.for.intercept = 10                           ,
                  prior.df.for.intercept    = 1                            ,
                  control                   = glmControl           
                ))
      #browser()
      options(warn=warn)
     #crossValLikelihoods[[crossValIndex, modelIndex]] = 
      crossValLikelihoods[ crossValIndex, modelIndex ] = 
        calcCrossValLikelihood(
                               coefficients        = myModels[[crossValIndex, modelIndex]]$coefficients, 
                               testSampleDat       = testSubsetRgen,  
                               testSampleOutcomes  = testSubsetccs  ,
                               trainSampleDat      = trainSubsetRgen,
                               trainSampleOutcomes = trainSubsetccs ,
                               simpleReturn        = TRUE           ,
                               lnullModelLike      = lnullModelLike
                              )
    }
  }
  if(verbose>1)
    cat(" Done\n")
 #cvls = apply(crossValLikelihoods, MARGIN=c(1,2), FUN='[', 1)
  cvls = crossValLikelihoods
  if(!crossValDetails[["replace"]]) {
    subModelMargLikes  = colSums(crossValLikelihoods)
    # Calculate the approximate marginal likelihoods and hence bayes factor(s)
    # I have: log(P(D | M, prior=priors[i])), and P(prior=priors[i]) for all i
    # I want: P(D | M) = sum_i P(D, prior=priors[i] | M) = sum_i P(D | M, prior=priors[i])P(prior=priors[i])
    # So, approach is to calculate:
    # log(P(D | M)) = AddLog_i (log(P(D, prior=priors[i] | M)))
    #               = AddLog_i (log(P(D | M, prior=priors[i])) + log(P(prior=priors[i])))
    # log(P(D | M, prior=priors[i])) = subModelMargLikes[i]
    # log(P(prior=priors[i]))        = models[["modelWeights"]][i]
    
    if(modelChoiceDetails[["weightModels"]]) {
      normModelWeights       = normalise(models[["modelWeights"]])
      weightedLikelihoodsVec = subModelMargLikes + log(normModelWeights)
      wholeModelMargLike     = addLog2(vec = weightedLikelihoodsVec)
    } else {
      wholeModelMargLike = addLog2(vec = subModelMargLikes) - log(sum(models[["modelsToUse"]]))
    }
  } else {
    cat("  Cross val with replacement not implemented.\n")
    browser()
  }
  if(is.null(names(subModelMargLikes)))
    names(subModelMargLikes) = models$modelNames[which(models$modelsToUse==1)]
  bayesFactor    = wholeModelMargLike - lnullModelLike
  rv = list("natLogBayesFactor"  = bayesFactor       ,
            "wholeModelMargLike" = wholeModelMargLike,
            "subModelMargLikes"  = subModelMargLikes ,
            "lnullModelLike"     = lnullModelLike
           )
  if(DEBUG>2 | verbose>2) {
    cat("  \n")
    cat("  wholeModelMargLike       = ", wholeModelMargLike                , "\n", sep="")
    cat("  lnullModelLike           = ", lnullModelLike                    , "\n", sep="")
    cat("  numInds                  = ", numInds                           , "\n", sep="")
    cat("  numSNPs                  = ", numSNPs                           , "\n", sep="")
    cat("  crossValFold             = ", crossValFold                      , "\n", sep="")
    cat("  numCrossVals             = ", numCrossVals                      , "\n", sep="")
    cat("  max(cor( gen, ccs))      = ", max(cor(caseControlStatus,  rGen)), "\n", sep="")
  }
  if(DEBUG>=2 | verbose>=1.75) {
    cat("  log10(bayesFactor)       = ", log10(exp(1))*bayesFactor         , "\n", sep="")
  }
  
  if(DEBUG>2) {
    if(abs(bayesFactor*log10(exp(1)))>(50/DEBUG+1) & DEBUG>1.5) {
      genPhenCor = cor(caseControlStatus, rGen)^2
      if(max(genPhenCor)<0.25 & bayesFactor*log10(exp(1))>50) {
        cat("  Warning, highly extreme log10 Bayes Factor (", bayesFactor*log10(exp(1)), ") \n", sep="")
        browser()
      }
    }
  }
  return(rv)  
}

